/******************************************************************************
*                                                                             *
* project name:    STARS                                                      *
* initial date:    06/06/2000                                                 *
* author:          thomas.nussbaumer@gmx.net                                  *
* description:     starfield effect demonstration (old version)               *
*                                                                             *
*******************************************************************************

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! NOTE: depends on TIGCC 0.93 and later to compile                            !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
-------------------------------------------------------------------------------
 Revision History                                                            
-------------------------------------------------------------------------------

v2.00: 12/02/2002
    by Lionel DEBROUX (lionel_debroux@yahoo.fr)

    Many modifications...
    Now it's a "true" screensaver (actually battery-waster, screensavers are
    *very* useless on LCDs).
    Set default contrast value to make the screensaver look better,
    whatever the contrast is. Anyway, you can change it with LEFT or RIGHT.
    Also, the speed of the starfield can be changed with ( and ). Finally, I
    managed to implement the 'backwards travel'...

    Credits go to Greg Dietsche.

    size optimization using -fomit-frame-pointer and static keyword for
    functions [tnussb]

v1.00: 16/06/2000
       initial version By Thomas NUSSBAUMER

*******************************************************************************/

#if defined(USE_TI89) && defined(USE_TI92P)
#error cannot produce both versions at once
#elif defined(USE_TI89)
#define C89_92(x,y) (x)
#elif defined(USE_TI92P)
#define C89_92(x,y) (y)
#else
#error Please specify either -DUSE_TI89 or -DUSE_TI92P on commandline
#endif

#define SAVE_SCREEN           // Save/Restore LCD Contents

#define OPTIMIZE_ROM_CALLS  	// Use ROM Call Optimization
#define NO_EXIT_SUPPORT

#include <tigcclib.h>

#define MY_LCD_WIDTH  C89_92(160,240)
#define MY_LCD_HEIGHT C89_92(100,128)


/*---------------------------------------------------------------------------*/
/* some constants defining the look of the starfield                         */
/*---------------------------------------------------------------------------*/
#define XOFFSET      (MY_LCD_WIDTH>>1)
#define YOFFSET      (MY_LCD_HEIGHT>>1)
#define ZOFFSET      200
#define MAX_STAR_X   16000
#define MAX_STAR_Y   16000
#define MAX_STARX2   (MAX_STAR_X>>1)
#define MAX_STARY2   (MAX_STAR_Y>>1)
#define BORDER_BLACK 100
#define BORDER_GRAY  140


/*---------------------------------------------------------------------------*/
/* structure which stores a star's location                                  */
/*---------------------------------------------------------------------------*/
typedef struct {
    short x,y,z;
    short offset;  // offset of old location for cleanup
    short mask;    // inverted mask of old location for cleanup
} star_type;


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
star_type* stars = NULL;  // array of stars

#define MAX_NUMBER_STARS        500   // number of maximum stars on screen
#define MIN_NUMBER_STARS        100   // number of minimum stars on screen
#define NUMBER_STARS_INCREMENT   20   // number of stars changed by each keypress

#define MAX_ZDECREMENT            9   // maximum "speed" in forward direction
#define MIN_ZDECREMENT           -9   // maximum "speed" in backward direction

static short number_stars = MAX_NUMBER_STARS;
static short zdecrement   = 4;


/*===========================================================================*/
/* initializes star with given index                                         */
/*===========================================================================*/
static void CreateStar(short index) {
    stars[index].x=(zdecrement>0)?random(MAX_STAR_X)-(MAX_STARX2):random(MAX_STAR_X+((MAX_STARX2)+1000))-((MAX_STAR_X+((MAX_STARX2)+1000))>>1);
    stars[index].y=(zdecrement>0)?random(MAX_STAR_X)-(MAX_STARY2):random(MAX_STAR_X+((MAX_STARX2)+1000))-((MAX_STAR_X+((MAX_STARX2)+1000))>>1);
    stars[index].z=(zdecrement>0)?ZOFFSET:random(ZOFFSET-1)+1;
}


/*===========================================================================*/
/* delete star at old position/calculate new position/draw new position      */
/*===========================================================================*/
static void DrawStars(void) {
    short         offset;
    unsigned char mask;
    short         drawx;
    short         drawy;
    short         i;

    for (i=0;i<number_stars;i++) {
        if (stars[i].offset != -1) {
            *(unsigned char *)(GetPlane(LIGHT_PLANE)+stars[i].offset) &= stars[i].mask;
            *(unsigned char *)(GetPlane(DARK_PLANE)+stars[i].offset) &= stars[i].mask;
            stars[i].offset = -1;
        }

        // calculate new drawing x position
        // and check if it out of screen
        drawx  = stars[i].x / stars[i].z + XOFFSET;
        if (drawx < 0 || drawx >= MY_LCD_WIDTH) {
            CreateStar(i);
            continue;
        }

        // calculate new drawing y position
        // and check if it out of screen
        drawy  = stars[i].y / stars[i].z + YOFFSET;
        if (drawy < 0 || drawy >= MY_LCD_HEIGHT) {
            CreateStar(i);
            continue;
        }

        // calculate byte offset of pixel
        offset = ((drawy<<5) - (drawy+drawy) + (drawx >> 3));

        // calculate mask value
        mask = 0x80 >> (drawx % 8);

        // set pixel in plane 1 and/or plane2
        // depending on z value
        if (stars[i].z < BORDER_BLACK) {
            *(unsigned char *)(GetPlane(LIGHT_PLANE)+offset) |= mask;
            *(unsigned char *)(GetPlane(DARK_PLANE)+offset)  |= mask;
        }

        else if (stars[i].z > BORDER_GRAY) {
            *(unsigned char *)(GetPlane(LIGHT_PLANE)+offset) |= mask;
        }

        else {
            *(unsigned char *)(GetPlane(DARK_PLANE)+offset) |= mask;
        }

        // store offset and mask so we can
        // easily remove this pixel later
        stars[i].mask   = ~mask;
        stars[i].offset = offset;

        // decrement z and check if we must
        // generate a new star
        stars[i].z -= zdecrement;
        if (stars[i].z <= 0 || stars[i].z>ZOFFSET) CreateStar(i);
    }
}


/*===========================================================================*/
/* clears 2 given planes                                                     */
/*===========================================================================*/
void ClearGrayScreen2B(void* lightplane,void* darkplane); // C prototype
asm("xdef ClearGrayScreen2B\n"
"ClearGrayScreen2B:\n"
"      movem.l  %d0-%d7/%a0-%a4,-(%a7)\n"
"      move.l   56(%a7),%a0\n"
"      move.l   60(%a7),%a1\n"
"      lea      (0xf00,%a0),%a0\n"
"      lea      (0xf00,%a1),%a1\n"
"      moveq    #23,%d7\n"
"      moveq    #0x0,%d0\n"
"      moveq    #0x0,%d1\n"
"      moveq    #0x0,%d2\n"
"      moveq    #0x0,%d3\n"
"      moveq    #0x0,%d4\n"
"      moveq    #0x0,%d5\n"
"      moveq    #0x0,%d6\n"
"      movea.l  %d0,%a2\n"
"      movea.l  %d0,%a3\n"
"      movea.l  %d0,%a4\n"
"0:\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a4,-(%a1)\n"
"      dbf      %d7, 0b\n"
"      movem.l  (%a7)+,%d0-%d7/%a0-%a4\n"
"      rts");


/*===========================================================================*/
/* nice macro to clear grayplanes using above clear routine                  */
/*===========================================================================*/
#define ClearGrayScreen() \
ClearGrayScreen2B(GetPlane(LIGHT_PLANE),GetPlane(DARK_PLANE))


/*===========================================================================*/
/* number of stars changed -> clear the screen and restart timer             */
/*===========================================================================*/
static void NrStarsChanged(void) {
    ClearGrayScreen();
    OSTimerRestart(APD_TIMER);
}


/*===========================================================================*/
/* Returns a pointer to the adress where is stored the contrast.             */
/* This function is from the famous TI-Chess, by Thomas Nussbaumer           */
/* and Marcos Lopez, via Franois Leiber's Othello II.                       */
/*===========================================================================*/
volatile unsigned char *contrast(void);
asm("contrast: move.w #4,%d0;trap #9;rts");


/*===========================================================================*/
/* wait a while for key debouncing                                           */
/*===========================================================================*/
static void debounce(void) {
    OSRegisterTimer(USER_TIMER,3);
    do {} while(!OSTimerExpired(USER_TIMER));
    OSFreeTimer(USER_TIMER);
    OSTimerRestart(APD_TIMER);
}

#define ARROWS_ROW       ~(1<<0)
#define ESC_ROW          ~C89_92(1<<6,1<<8)
#define OPEN_PARENTH_ROW ~(1<<4)
#define CLOS_PARENTH_ROW ~C89_92(1<<3,1<<4)

#define UP           C89_92(1<<0,1<<5)
#define DOWN         C89_92(1<<2,1<<7)
#define LEFT         C89_92(1<<1,1<<4)
#define RIGHT        C89_92(1<<3,1<<6)
#define ESC          C89_92(1<<0,1<<6)
#define OPEN_PARENTH C89_92(1<<4,1<<5)
#define CLOS_PARENTH C89_92(1<<4,1<<6)


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    short          i;
    volatile short contsave = 0;
    INT_HANDLER    save_int1;

    // show help screen
    ClrScr();
    FontSetSys(F_8x10);
    DrawStr(C89_92(15,55),15,"Starfield Effect",A_NORMAL);
    DrawStr(C89_92(25,65),30,"(c) TICT 2002",A_NORMAL);
    FontSetSys(F_4x6);
    DrawStr(C89_92(15,55),50,"UP / DOWN        modify number of stars",A_NORMAL);
    DrawStr(C89_92(15,55),58,"LEFT / RIGHT    modify constrast",A_NORMAL);
    DrawStr(C89_92(15,55),66,"( / )                   modify speed/direction",A_NORMAL);
    ngetchx();

    // redirect default interrupt-1 handler to dummy handler
    save_int1=GetIntVec(AUTO_INT_1);
    SetIntVec(AUTO_INT_1,DUMMY_HANDLER);

    // turn on grayscales
    if (!GrayOn()) {
        SetIntVec(AUTO_INT_1,save_int1);
        return;
    }

    // try to allocate memory for stars
    if(!(stars=(star_type*)malloc(sizeof(star_type)*MAX_NUMBER_STARS))) {
        GrayOff();
        SetIntVec(AUTO_INT_1,save_int1);
        ClrScr();
        DrawStr(0,0,"ERROR: not enough memory",A_REPLACE);
        GKeyIn(NULL,0);
        return;
    }

    // setup initial star positions
    for (i=0;i<number_stars;i++) {
        stars[i].x=random(MAX_STAR_X)-(MAX_STAR_X>>1);
        stars[i].y=random(MAX_STAR_X)-(MAX_STAR_Y>>1);
        stars[i].z=random(ZOFFSET)+1;
        stars[i].offset = -1;
    }

    // clear both planes
    ClearGrayScreen();

    // get initial contrast
    contsave=*contrast();

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    while (1) {
        DrawStars();
        if (_rowread(ARROWS_ROW)&UP && (number_stars<=(MAX_NUMBER_STARS-NUMBER_STARS_INCREMENT))) {  // UP
            number_stars+=NUMBER_STARS_INCREMENT;
            debounce();
            NrStarsChanged();
        }
        else if (_rowread(ARROWS_ROW)&DOWN && (number_stars>=MIN_NUMBER_STARS+NUMBER_STARS_INCREMENT)) {  // DOWN
            number_stars-=NUMBER_STARS_INCREMENT;
            debounce();
            NrStarsChanged();
        }
        else if (_rowread(ARROWS_ROW)&RIGHT) {  // RIGHT
            OSContrastUp();
            debounce();
        }
        else if (_rowread(ARROWS_ROW)&LEFT) {  // LEFT
            OSContrastDn();
            debounce();
        }
        else if (_rowread(OPEN_PARENTH_ROW)&OPEN_PARENTH && (zdecrement>MIN_ZDECREMENT)) {  // (
            if (!(zdecrement-1)) zdecrement=-1;
            else zdecrement--;
            debounce();
        }
        else if (_rowread(CLOS_PARENTH_ROW)&CLOS_PARENTH && (zdecrement<MAX_ZDECREMENT)) {  // )
            if (!(zdecrement+1)) zdecrement=1;
            else zdecrement++;
            debounce();
        }
        else if (_rowread(ESC_ROW)&ESC) {  // ESC
            *contrast()=contsave;
            OSContrastDn();  // strange contrast will be set on VTI back only with
            OSContrastUp();  // this down/up combo !?!?
            ClearGrayScreen();
            GKeyFlush();
            while(!(_rowread(0xF000)&0xFF));
            GKeyFlush();
            debounce();

            // cleanup and bye =8-)
            free(stars);
            GKeyFlush();
            GrayOff();
            SetIntVec(AUTO_INT_1,save_int1);
            FontSetSys(F_6x8);
            return;
        }
        if (OSTimerExpired(APD_TIMER)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
    }
}
